"""Provide the Subreddits class."""
from .base import PRAWBase
from .listing.generator import ListingGenerator
from .util import stream_generator
from ..const import API_PATH


class Subreddits(PRAWBase):
    """Subreddits is a Listing class that provides various subreddit lists."""

    def default(self, **generator_kwargs):
        """Return a ListingGenerator for default subreddits."""
        return ListingGenerator(self._reddit, API_PATH['subreddits_default'],
                                **generator_kwargs)

    def gold(self, **generator_kwargs):
        """Return a ListingGenerator for gold subreddits."""
        return ListingGenerator(self._reddit, API_PATH['subreddits_gold'],
                                **generator_kwargs)

    def new(self, **generator_kwargs):
        """Return a ListingGenerator for new subreddits."""
        return ListingGenerator(self._reddit, API_PATH['subreddits_new'],
                                **generator_kwargs)

    def popular(self, **generator_kwargs):
        """Return a ListingGenerator for default subreddits."""
        return ListingGenerator(self._reddit, API_PATH['subreddits_popular'],
                                **generator_kwargs)

    def search(self, query, **generator_kwargs):
        """Return a ListingGenerator of subreddits matching ``query``.

        Subreddits are searched by both their title and description. To search
        names only see ``search_by_name``.

        :param query: The query string to filter subreddits by.

        """
        generator_kwargs.setdefault('params', {})['q'] = query
        return ListingGenerator(self._reddit, API_PATH['subreddits_search'],
                                **generator_kwargs)

    def search_by_name(self, query, include_nsfw=True, exact=False):
        """Return list of Subreddits whose names begin with ``query``.

        :param query: Search for subreddits beginning with this string.
        :param include_nsfw: Include subreddits labeled NSFW (default: True).
        :param exact: Return only exact matches to ``query`` (default: False).

        """
        result = self._reddit.post(API_PATH['subreddits_name_search'],
                                   data={'include_over_18': include_nsfw,
                                         'exact': exact, 'query': query})
        return [self._reddit.subreddit(x) for x in result['names']]

    def stream(self):
        """Yield new subreddits as they are created.

        Subreddits are yielded oldest first. Up to 100 historial subreddits
        will initially be returned.

        """
        return stream_generator(self.new)



from urllib.parse import quote_plus

import praw

QUESTIONS = ['what is', 'who is', 'what are']
REPLY_TEMPLATE = '[Let me google that for you](http://lmgtfy.com/?q={})'


def main():
    reddit = praw.Reddit(user_agent='LMGTFY (by /u/USERNAME)',
                         client_id='CLIENT_ID', client_secret="CLIENT_SECRET",
                         username='USERNAME', password='PASSWORD')

    subreddit = reddit.subreddit('AskReddit')
    for submission in subreddit.stream.submissions():
        process_submission(submission)


def process_submission(submission):
    # Ignore titles with more than 10 words as they probably are not simple
    # questions.
    if len(submission.title.split()) > 10:
        return

    normalized_title = submission.title.lower()
    for question_phrase in QUESTIONS:
        if question_phrase in normalized_title:
            url_title = quote_plus(submission.title)
            reply_text = REPLY_TEMPLATE.format(url_title)
            print('Replying to: {}'.format(submission.title))
            submission.reply(reply_text)
            # A reply has been made so do not attempt to match other phrases.
            break


if __name__ == '__main__':
    main()



"""Test praw.models.front."""
from .. import IntegrationTest


class TestFront(IntegrationTest):
    def test_controversial(self):
        with self.recorder.use_cassette('TestFront.test_controversial'):
            submissions = list(self.reddit.front.controversial())
        assert len(submissions) == 100

    def test_gilded(self):
        with self.recorder.use_cassette('TestFront.test_gilded'):
            submissions = list(self.reddit.front.gilded())
        assert len(submissions) == 100

    def test_hot(self):
        with self.recorder.use_cassette('TestFront.test_hot'):
            submissions = list(self.reddit.front.hot())
        assert len(submissions) == 100

    def test_new(self):
        with self.recorder.use_cassette('TestFront.test_new'):
            submissions = list(self.reddit.front.new())
        assert len(submissions) == 100

    def test_top(self):
        with self.recorder.use_cassette('TestFront.test_top'):
            submissions = list(self.reddit.front.top())
        assert len(submissions) == 100



"""Test praw.models.auth."""
from praw import Reddit
from praw.exceptions import ClientException
import pytest

from .. import UnitTest


class TestAuth(UnitTest):
    def test_implicit__from_script_app(self):
        with pytest.raises(ClientException):
            self.reddit.auth.implicit('dummy token', 10, '')

    def test_url__installed_app(self):
        reddit = Reddit(client_id='dummy client', client_secret=None,
                        redirect_uri='https://dummy.tld/',
                        user_agent='dummy')
        url = reddit.auth.url(['dummy scope'], 'dummy state')
        assert 'client_id=dummy+client' in url
        assert 'redirect_uri=https%3A%2F%2Fdummy.tld%2F' in url
        assert 'scope=dummy+scope' in url
        assert 'state=dummy+state' in url

    def test_url__unsupported_app(self):
        with pytest.raises(ClientException):
            self.reddit.auth.url('dummy scope', 'dummy state')



"""Test praw.models.list.base."""
import pytest

from praw.models.list.base import BaseList


class TestBaseList(object):
    def setup(self):
        self._prev_child_attribute = BaseList.CHILD_ATTRIBUTE
        self._prev_convert = BaseList._convert

    def teardown(self):
        BaseList.CHILD_ATTRIBUTE = self._prev_child_attribute
        BaseList._convert = staticmethod(self._prev_convert)

    def test__init__CHILD_ATTRIBUTE_not_set(self):
        with pytest.raises(NotImplementedError):
            BaseList(None, None)

    def test__init___convert_not_extended(self):
        BaseList.CHILD_ATTRIBUTE = 'praw'
        with pytest.raises(NotImplementedError):
            BaseList(None, {'praw': [1]})

    def test__contains__(self):
        BaseList._convert = staticmethod(lambda _a, _b: None)
        BaseList.CHILD_ATTRIBUTE = 'praw'
        items = ['foo', 1, {'a': 'b'}]
        base_list = BaseList(None, {'praw': items})
        for item in items:
            assert item in base_list

    def test__getitem__(self):
        BaseList._convert = staticmethod(lambda _a, _b: None)
        BaseList.CHILD_ATTRIBUTE = 'praw'
        items = ['foo', 1, {'a': 'b'}]
        base_list = BaseList(None, {'praw': items})
        for i, item in enumerate(items):
            assert item == base_list[i]

    def test__iter__(self):
        BaseList._convert = staticmethod(lambda _a, _b: None)
        BaseList.CHILD_ATTRIBUTE = 'praw'
        items = ['foo', 1, {'a': 'b'}]
        base_list = BaseList(None, {'praw': items})
        for i, item in enumerate(base_list):
            assert items[i] == item

    def test__str__(self):
        BaseList._convert = staticmethod(lambda _a, _b: None)
        BaseList.CHILD_ATTRIBUTE = 'praw'
        items = ['foo', 1, {'a': 'b'}]
        base_list = BaseList(None, {'praw': items})
        assert str(items) == str(base_list)



"""Provide the Message class."""
from .base import RedditBase
from ...const import API_PATH
from .mixins import InboxableMixin


class Message(RedditBase, InboxableMixin):
    """A class for private messages."""

    @classmethod
    def parse(cls, data, reddit):
        """Return an instance of Message or SubredditMessage from ``data``.

        :param data: The structured data.
        :param reddit: An instance of :class:`.Reddit`.

        """
        if data.get('subreddit'):
            return SubredditMessage(reddit, _data=data)
        return cls(reddit, _data=data)

    def __init__(self, reddit, _data):
        """Construct an instance of the Message object."""
        super(Message, self).__init__(reddit, _data)


class SubredditMessage(Message):
    """A class for messages to a subreddit."""

    def mute(self, _unmute=False):
        """Mute the sender of this SubredditMessage."""
        self._reddit.post(API_PATH['mute_sender'], data={'id': self.fullname})

    def unmute(self):
        """Unmute the sender of this SubredditMessage."""
        self._reddit.post(API_PATH['unmute_sender'],
                          data={'id': self.fullname})



"""Test praw.models.subreddit."""
from praw.exceptions import APIException
from praw.models import (Comment, Redditor, Submission, SubredditMessage,
                         WikiPage)
import mock
import pytest

from ... import IntegrationTest


class TestSubreddit(IntegrationTest):
    @mock.patch('time.sleep', return_value=None)
    def test_create(self, _):
        self.reddit.read_only = False
        new_name = 'PRAW_rrldkyrfln'
        with self.recorder.use_cassette('TestSubreddit.test_create'):
            subreddit = self.reddit.subreddit.create(name=new_name,
                                                     title='Sub',
                                                     link_type='any',
                                                     subreddit_type='public',
                                                     wikimode='disabled')
            assert subreddit.display_name == new_name
            assert subreddit.submission_type == 'any'

    def test_create__exists(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubreddit.test_create__exists'):
            with pytest.raises(APIException) as excinfo:
                self.reddit.subreddit.create('redditdev', title='redditdev',
                                             link_type='any',
                                             subreddit_type='public',
                                             wikimode='disabled')
            assert excinfo.value.error_type == 'SUBREDDIT_EXISTS'

    def test_create__invalid_parameter(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubreddit.test_create__invalid_parameter'):
            with pytest.raises(APIException) as excinfo:
                # Supplying invalid setting for link_type
                self.reddit.subreddit.create(name='PRAW_iavynavffv',
                                             title='sub', link_type='abcd',
                                             subreddit_type='public',
                                             wikimode='disabled')
            assert excinfo.value.error_type == 'INVALID_OPTION'

    def test_create__missing_parameter(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubreddit.test_create__missing_parameter'):
            with pytest.raises(APIException) as excinfo:
                # Not supplying required field title.
                self.reddit.subreddit.create(name='PRAW_iavynavffv',
                                             title=None, link_type='any',
                                             subreddit_type='public',
                                             wikimode='disabled')
            assert excinfo.value.error_type == 'NO_TEXT'

    @mock.patch('time.sleep', return_value=None)
    def test_message(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubreddit.test_message'):
            subreddit = self.reddit.subreddit(
                pytest.placeholders.test_subreddit)
            subreddit.message('Test from PRAW', message='Test content')

    def test_random(self):
        with self.recorder.use_cassette('TestSubreddit.test_random'):
            subreddit = self.reddit.subreddit(
                pytest.placeholders.test_subreddit)
            submissions = [subreddit.random(), subreddit.random(),
                           subreddit.random(), subreddit.random()]
            assert len(submissions) == len(set(submissions))

    def test_search(self):
        with self.recorder.use_cassette('TestSubreddit.test_search'):
            subreddit = self.reddit.subreddit('all')
            for item in subreddit.search('praw oauth search', limit=None):
                assert isinstance(item, Submission)

    @mock.patch('time.sleep', return_value=None)
    def test_submit__selftext(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubreddit.test_submit__selftext'):
            subreddit = self.reddit.subreddit(
                pytest.placeholders.test_subreddit)
            submission = subreddit.submit('Test Title', selftext='Test text.')
            assert submission.author == self.reddit.config.username
            assert submission.selftext == 'Test text.'
            assert submission.title == 'Test Title'

    @mock.patch('time.sleep', return_value=None)
    def test_submit__url(self, _):
        url = 'https://praw.readthedocs.org/en/stable/'
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubreddit.test_submit__url'):
            subreddit = self.reddit.subreddit(
                pytest.placeholders.test_subreddit)
            submission = subreddit.submit('Test Title', url=url)
            assert submission.author == self.reddit.config.username
            assert submission.url == url
            assert submission.title == 'Test Title'


class TestSubredditFlair(IntegrationTest):
    @property
    def subreddit(self):
        return self.reddit.subreddit(pytest.placeholders.test_subreddit)

    def test__iter(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubredditFlair.test__iter'):
            mapping = list(self.subreddit.flair)
            assert len(mapping) > 0
            assert all(isinstance(x['user'], Redditor) for x in mapping)

    @mock.patch('time.sleep', return_value=None)
    def test_delete_all(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubredditFlair.test_delete_all'):
            response = self.subreddit.flair.delete_all()
            assert len(response) > 100
            assert all('removed' in x['status'] for x in response)

    def test_set__redditor(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubredditFlair.test_set__redditor'):
            redditor = self.subreddit._reddit.redditor(
                self.reddit.config.username)
            self.subreddit.flair.set(redditor, 'redditor flair')

    def test_set__redditor_string(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubredditFlair.test_set__redditor_string'):
            self.subreddit.flair.set(self.reddit.config.username, 'new flair',
                                     'some class')

    def test_set__submission(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubredditFlair.test_set__submission'):
            submission = self.subreddit._reddit.submission('4b536p')
            self.subreddit.flair.set(submission, 'submission flair')

    def test_update(self):
        self.reddit.read_only = False
        redditor = self.subreddit._reddit.redditor(self.reddit.config.username)
        flair_list = [redditor, 'spez', {'user': 'bsimpson'},
                      {'user': 'spladug', 'flair_text': '',
                       'flair_css_class': ''}]
        with self.recorder.use_cassette('TestSubredditFlair.test_update'):
            response = self.subreddit.flair.update(flair_list,
                                                   css_class='default')
        assert all(x['ok'] for x in response)
        assert not any(x['errors'] for x in response)
        assert not any(x['warnings'] for x in response)
        assert len([x for x in response if 'added' in x['status']]) == 3
        assert len([x for x in response if 'removed' in x['status']]) == 1
        for i, name in enumerate([str(redditor), 'spez', 'bsimpson',
                                  'spladug']):
            assert name in response[i]['status']

    def test_update__comma_in_text(self):
        self.reddit.read_only = False
        flair_list = [{'user': 'bsimpson'},
                      {'user': 'spladug', 'flair_text': 'a,b'}]
        with self.recorder.use_cassette(
                'TestSubredditFlair.test_update__comma_in_text'):
            response = self.subreddit.flair.update(flair_list,
                                                   css_class='default')
        assert all(x['ok'] for x in response)


class TestSubredditListings(IntegrationTest):
    def test_comments(self):
        with self.recorder.use_cassette(
                'TestSubredditListings.test_comments'):
            subreddit = self.reddit.subreddit('askreddit')
            comments = list(subreddit.comments())
        assert len(comments) == 100

    def test_comments_gilded(self):
        with self.recorder.use_cassette(
                'TestSubredditListings.test_comments_gilded'):
            subreddit = self.reddit.subreddit('askreddit')
            comments = list(subreddit.comments.gilded())
        assert len(comments) == 100

    def test_controversial(self):
        with self.recorder.use_cassette(
                'TestSubredditListings.test_controversial'):
            subreddit = self.reddit.subreddit('askreddit')
            submissions = list(subreddit.controversial())
        assert len(submissions) == 100

    def test_gilded(self):
        with self.recorder.use_cassette('TestSubredditListings.test_gilded'):
            subreddit = self.reddit.subreddit('askreddit')
            submissions = list(subreddit.gilded())
        assert len(submissions) >= 50

    def test_hot(self):
        with self.recorder.use_cassette('TestSubredditListings.test_hot'):
            subreddit = self.reddit.subreddit('askreddit')
            submissions = list(subreddit.hot())
        assert len(submissions) == 100

    def test_new(self):
        with self.recorder.use_cassette('TestSubredditListings.test_new'):
            subreddit = self.reddit.subreddit('askreddit')
            submissions = list(subreddit.new())
        assert len(submissions) == 100

    def test_rising(self):
        with self.recorder.use_cassette('TestSubredditListings.test_rising'):
            subreddit = self.reddit.subreddit('askreddit')
            submissions = list(subreddit.rising())
        assert len(submissions) == 100

    def test_top(self):
        with self.recorder.use_cassette('TestSubredditListings.test_top'):
            subreddit = self.reddit.subreddit('askreddit')
            submissions = list(subreddit.top())
        assert len(submissions) == 100


class TestSubredditModeration(IntegrationTest):
    @property
    def subreddit(self):
        return self.reddit.subreddit(pytest.placeholders.test_subreddit)

    def test_approve(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubredditModeration.test_approve'):
            submission = self.reddit.submission('4b536h')
            self.subreddit.mod.approve(submission)

    def test_distinguish(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubredditModeration.test_distinguish'):
            submission = self.reddit.submission('4b536h')
            self.subreddit.mod.distinguish(submission)

    def test_ignore_reports(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubredditModeration.test_ignore_reports'):
            submission = self.reddit.submission('31ybt2')
            self.subreddit.mod.ignore_reports(submission)

    def test_inbox(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubredditModeration.test_inbox'):
            count = 0
            for item in self.reddit.subreddit('all').mod.inbox():
                assert isinstance(item, SubredditMessage)
                count += 1
            assert count == 100

    def test_remove(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubredditModeration.test_remove'):
            submission = self.reddit.submission('4b536h')
            self.subreddit.mod.remove(submission, spam=True)

    def test_undistinguish(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubredditModeration.test_undistinguish'):
            submission = self.reddit.submission('4b536h')
            self.subreddit.mod.undistinguish(submission)

    def test_unignore_reports(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubredditModeration.test_unignore_reports'):
            submission = self.reddit.submission('31ybt2')
            self.subreddit.mod.unignore_reports(submission)

    def test_unread(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubredditModeration.test_unread'):
            count = 0
            for item in self.reddit.subreddit('all').mod.unread():
                assert isinstance(item, SubredditMessage)
                count += 1
            assert count > 0

    @mock.patch('time.sleep', return_value=None)
    def test_update(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubredditModeration.test_update'):
            before_settings = self.subreddit.mod.settings()
            new_title = before_settings['title'] + 'x'
            if len(new_title) == 20:
                new_title = 'x'
            self.subreddit.mod.update(title=new_title)
            assert self.subreddit.title == new_title
            after_settings = self.subreddit.mod.settings()

            # Ensure that nothing has changed besides what was specified.
            before_settings['title'] = new_title
            assert before_settings == after_settings


class TestSubredditRelationships(IntegrationTest):
    REDDITOR = 'pyapitestuser3'

    @mock.patch('time.sleep', return_value=None)
    def add_remove(self, subreddit, user, relationship, _):
        relationship = getattr(subreddit, relationship)
        relationship.add(user)
        assert user in relationship
        relationship.remove(user)
        assert user not in relationship

    @property
    def subreddit(self):
        return self.reddit.subreddit(pytest.placeholders.test_subreddit)

    def test_banned(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubredditRelationships__banned'):
            self.add_remove(self.subreddit, self.REDDITOR, 'banned')

    def test_contributor(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubredditRelationships__contributor'):
            self.add_remove(self.subreddit, self.REDDITOR, 'contributor')

    @mock.patch('time.sleep', return_value=None)
    def test_moderator(self, _):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubredditRelationships__moderator'):
            # Moderators can only be invited.
            # As of 2016-03-18 there is no API endpoint to get the moderator
            # invite list.
            self.subreddit.moderator.add(self.REDDITOR)
            assert self.REDDITOR not in self.subreddit.moderator

    def test_muted(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette('TestSubredditRelationships__muted'):
            self.add_remove(self.subreddit, self.REDDITOR, 'muted')

    def test_wiki_banned(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubredditRelationships__wiki_banned'):
            self.add_remove(self.subreddit.wiki, self.REDDITOR, 'banned')

    def test_wiki_contributors(self):
        self.reddit.read_only = False
        with self.recorder.use_cassette(
                'TestSubredditRelationships__wiki_contributors'):
            self.add_remove(self.subreddit.wiki, self.REDDITOR, 'contributors')


class TestSubredditStreams(IntegrationTest):
    def test_comments(self):
        with self.recorder.use_cassette('TestSubredditStreams__comments'):
            subreddit = self.reddit.subreddit(
                pytest.placeholders.test_subreddit)
            generator = subreddit.stream.comments()
            for i in range(101):
                assert isinstance(next(generator), Comment)

    def test_submissions(self):
        with self.recorder.use_cassette('TestSubredditStreams__submissions'):
            generator = self.reddit.subreddit('all').stream.submissions()
            for i in range(300):
                assert isinstance(next(generator), Submission)


class TestSubredditWiki(IntegrationTest):
    @mock.patch('time.sleep', return_value=None)
    def test__iter(self, _):
        self.reddit.read_only = False
        subreddit = self.reddit.subreddit(
            pytest.placeholders.test_subreddit)
        with self.recorder.use_cassette('TestSubredditWiki__iter'):
            count = 0
            for wikipage in subreddit.wiki:
                assert isinstance(wikipage, WikiPage)
                count += 1
            assert count > 0

    @mock.patch('time.sleep', return_value=None)
    def test_create(self, _):
        self.reddit.read_only = False
        subreddit = self.reddit.subreddit(
            pytest.placeholders.test_subreddit)

        with self.recorder.use_cassette('TestSubredditWiki_create'):
            wikipage = subreddit.wiki.create('PRAW New Page',
                                             'This is the new wiki page')
            assert wikipage.name == 'praw_new_page'
            assert wikipage.content_md == 'This is the new wiki page'



"""PRAW Unit test suite."""
from praw import Reddit


class UnitTest(object):
    """Base class for PRAW unit tests."""

    def setup(self):
        """Setup runs before all test cases."""
        self.reddit = Reddit(client_id='dummy', client_secret='dummy',
                             user_agent='dummy')
        # Unit tests should never issue requests
        self.reddit._core._requestor._http = None



"""Provide the BaseListingMixin class."""
from six.moves.urllib.parse import urljoin  # pylint: disable=import-error

from ...base import PRAWBase
from ..generator import ListingGenerator


def _prepare(praw_object, params, target):
    """Fix for Redditor methods that use a query param rather than subpath."""
    if praw_object.__dict__.get('_listing_use_sort'):
        params['sort'] = target
        return praw_object._path
    return urljoin(praw_object._path, target)


class BaseListingMixin(PRAWBase):
    """Adds minimum set of methods that apply to all listing objects."""

    VALID_TIME_FILTERS = {'all', 'day', 'hour', 'month', 'week', 'year'}

    @staticmethod
    def validate_time_filter(time_filter):
        """Raise ValueError if time_filter is not valid."""
        if time_filter not in BaseListingMixin.VALID_TIME_FILTERS:
            raise ValueError('time_filter must be one of: {}'.format(', '.join(
                BaseListingMixin.VALID_TIME_FILTERS)))

    def controversial(self, time_filter='all', **generator_kwargs):
        """Return a ListingGenerator for controversial submissions.

        :param time_filter: Can be one of: all, day, hour, month, week, year.
            (Default: all)

        Raise ``ValueError`` if ``time_filter`` is invalid.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        self.validate_time_filter(time_filter)
        generator_kwargs.setdefault('params', {})['t'] = time_filter
        url = _prepare(self, generator_kwargs['params'], 'controversial')
        return ListingGenerator(self._reddit, url, **generator_kwargs)

    def hot(self, **generator_kwargs):
        """Return a ListingGenerator for hot items.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        url = _prepare(self, generator_kwargs.setdefault('params', {}), 'hot')
        return ListingGenerator(self._reddit, url, **generator_kwargs)

    def new(self, **generator_kwargs):
        """Return a ListingGenerator for new items.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        url = _prepare(self, generator_kwargs.setdefault('params', {}), 'new')
        return ListingGenerator(self._reddit, url, **generator_kwargs)

    def top(self, time_filter='all', **generator_kwargs):
        """Return a ListingGenerator for top submissions.

        :param time_filter: Can be one of: all, day, hour, month, week, year.
            (Default: all)

        Raise ``ValueError`` if ``time_filter`` is invalid.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        self.validate_time_filter(time_filter)
        generator_kwargs.setdefault('params', {})['t'] = time_filter
        url = _prepare(self, generator_kwargs['params'], 'top')
        return ListingGenerator(self._reddit, url, **generator_kwargs)



import pickle

from praw.models import Subreddit, WikiPage

from ... import UnitTest


class TestWikiPage(UnitTest):
    def test_equality(self):
        page1 = WikiPage(self.reddit, subreddit=Subreddit(self.reddit, 'a'),
                         name='x')
        page2 = WikiPage(self.reddit, subreddit=Subreddit(self.reddit, 'a'),
                         name='2')
        page3 = WikiPage(self.reddit, subreddit=Subreddit(self.reddit, 'b'),
                         name='1')
        page4 = WikiPage(self.reddit, subreddit=Subreddit(self.reddit, 'A'),
                         name='x')
        page5 = WikiPage(self.reddit, subreddit=Subreddit(self.reddit, 'a'),
                         name='X')
        assert page1 == page1
        assert page2 == page2
        assert page3 == page3
        assert page1 != page2
        assert page1 != page3
        assert page1 == page4
        assert page1 == page5

    def test_hash(self):
        page1 = WikiPage(self.reddit, subreddit=Subreddit(self.reddit, 'a'),
                         name='x')
        page2 = WikiPage(self.reddit, subreddit=Subreddit(self.reddit, 'a'),
                         name='2')
        page3 = WikiPage(self.reddit, subreddit=Subreddit(self.reddit, 'b'),
                         name='1')
        page4 = WikiPage(self.reddit, subreddit=Subreddit(self.reddit, 'A'),
                         name='x')
        page5 = WikiPage(self.reddit, subreddit=Subreddit(self.reddit, 'a'),
                         name='X')
        assert hash(page1) == hash(page1)
        assert hash(page2) == hash(page2)
        assert hash(page3) == hash(page3)
        assert hash(page1) != hash(page2)
        assert hash(page1) != hash(page3)
        assert hash(page1) == hash(page4)
        assert hash(page1) == hash(page5)

    def test_pickle(self):
        page = WikiPage(self.reddit, subreddit=Subreddit(self.reddit, 'a'),
                        name='x')
        for level in range(pickle.HIGHEST_PROTOCOL + 1):
            other = pickle.loads(pickle.dumps(page, protocol=level))
            assert page == other

    def test_repr(self):
        page = WikiPage(self.reddit, subreddit=Subreddit(self.reddit, 'a'),
                        name='x')
        assert repr(page) == ('WikiPage(subreddit=Subreddit(display_name=\'a\''
                              '), name=\'x\')')

    def test_str(self):
        page = WikiPage(self.reddit, subreddit=Subreddit(self.reddit, 'a'),
                        name='x')
        assert str(page) == 'a/x'



"""Test praw.models.listing.generator."""
from ... import IntegrationTest


class TestListingGenerator(IntegrationTest):
    def test_exhaust_items(self):
        with self.recorder.use_cassette(
                'TestListingGenerator.test_exhaust_items'):
            submissions = list(self.reddit.redditor('spez').top(limit=None))
        assert len(submissions) > 100

    def test_no_items(self):
        with self.recorder.use_cassette('TestListingGenerator.test_no_items'):
            submissions = list(self.reddit.redditor('spez').top('hour'))
        assert len(submissions) == 0



"""Provide the RedditorList class."""
from ..reddit.redditor import Redditor
from .base import BaseList


class RedditorList(BaseList):
    """A list of Redditors. Works just like a regular list."""

    CHILD_ATTRIBUTE = 'children'

    @staticmethod
    def _convert(reddit, data):
        """Return a Redditor object from the data."""
        return Redditor(reddit, data['name'])



"""Provide the InboxableMixin class."""

from ....const import API_PATH


class InboxableMixin(object):
    """Interface for RedditBase classes that originate from the inbox."""

    def block(self):
        """Block the user who sent the item.

        :returns: The json response from the server.

        Note: reddit does not permit blocking users unless they you have an
        Inboxable item from them.

        """
        return self._reddit.post(API_PATH['block'], data={'id': self.fullname})

    def mark_read(self):
        """Mark the item as read."""
        self._reddit.inbox.mark_read([self])

    def mark_unread(self):
        """Mark the item as unread."""
        self._reddit.inbox.mark_unread([self])



"""Test praw.models.subreddits."""
from praw.models import Subreddit

from .. import IntegrationTest


class TestSubreddits(IntegrationTest):
    def test_default(self):
        with self.recorder.use_cassette('TestSubreddits.test_default'):
            subreddits = list(self.reddit.subreddits.default(limit=None))
        assert 0 < len(subreddits) < 100

    def test_gold__without_gold(self):
        with self.recorder.use_cassette(
                'TestSubreddits.test_gold__without_gold'):
            subreddits = list(self.reddit.subreddits.gold())
        assert len(subreddits) == 0

    def test_new(self):
        with self.recorder.use_cassette('TestSubreddits.test_new'):
            subreddits = list(self.reddit.subreddits.new(limit=300))
        assert len(subreddits) == 300

    def test_popular(self):
        with self.recorder.use_cassette('TestSubreddits.test_popular'):
            subreddits = list(self.reddit.subreddits.popular())
        assert len(subreddits) == 100

    def test_search(self):
        with self.recorder.use_cassette('TestSubreddits.test_search'):
            found = False
            for subreddit in self.reddit.subreddits.search('praw'):
                assert isinstance(subreddit, Subreddit)
                found = True
            assert found

    def test_search_by_name(self):
        with self.recorder.use_cassette('TestSubreddits.test_search_by_name'):
            subreddits = self.reddit.subreddits.search_by_name('reddit')
            assert isinstance(subreddits, list)
            assert len(subreddits) > 1
            assert all(isinstance(x, Subreddit) for x in subreddits)

    def test_stream(self):
        with self.recorder.use_cassette(
                'TestSubreddits__test_streams'):
            generator = self.reddit.subreddits.stream()
            for i in range(101):
                assert isinstance(next(generator), Subreddit)



"""praw setup.py"""

import re
from codecs import open
from os import path
from setuptools import find_packages, setup


PACKAGE_NAME = 'praw'
HERE = path.abspath(path.dirname(__file__))
with open(path.join(HERE, 'README.rst'), encoding='utf-8') as fp:
    README = fp.read()
with open(path.join(HERE, PACKAGE_NAME, 'const.py'),
          encoding='utf-8') as fp:
    VERSION = re.search("__version__ = '([^']+)'", fp.read()).group(1)


setup(name=PACKAGE_NAME,
      author='Bryce Boe',
      author_email='bbzbryce@gmail.com',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: Implementation :: CPython',
          'Topic :: Utilities'],
      description=('PRAW, an acronym for `Python Reddit API Wrapper`, is a '
                   'python package that allows for simple access to '
                   'reddit\'s API.'),
      install_requires=['decorator >=4.0.9, <4.1',
                        'prawcore >=0.2.1, <0.3',
                        'requests >=2.3.0',
                        'six ==1.10',
                        'update_checker ==0.11'],
      keywords='reddit api wrapper',
      license='GPLv3',
      long_description=README,
      package_data={'': ['COPYING'], PACKAGE_NAME: ['*.ini']},
      packages=find_packages(exclude=['tests', 'tests.*']),
      setup_requires=['pytest-runner ==2.7'],
      tests_require=['betamax >=0.7.1, <0.8',
                     'betamax-matchers >=0.2.0, <0.3',
                     'betamax-serializers >=0.2, <0.3',
                     'mock ==1.0.1',
                     'pytest ==2.8.7'],
      test_suite='tests',
      url='https://praw.readthedocs.org/',
      version=VERSION)



"""Package providing models that pertain to listing mixins."""
from .redditor import RedditorListingMixin  # NOQA
from .submission import SubmissionListingMixin  # NOQA
from .subreddit import SubredditListingMixin  # NOQA



"""Test praw.models.comment_forest."""
from praw.models import Submission

from .. import IntegrationTest


class TestCommentForest(IntegrationTest):
    def setup(self):
        super(TestCommentForest, self).setup()
        # Responses do not decode well on travis so manually renable gzip.
        self.reddit._core._requestor._http.headers['Accept-Encoding'] = 'gzip'

    def test_replace__all(self):
        with self.recorder.use_cassette(
                'TestCommentForest.test_replace__all',
                match_requests_on=['uri', 'method', 'body']):
            submission = Submission(self.reddit, '3hahrw')
            before_count = len(submission.comments.list())
            skipped = submission.comments.replace_more(None, threshold=0)
            assert len(skipped) == 0
            assert all(x.submission == submission for x in
                       submission.comments.list())
            assert before_count < len(submission.comments.list())

    def test_replace__all_large(self):
        with self.recorder.use_cassette(
                'TestCommentForest.test_replace__all_large',
                match_requests_on=['uri', 'method', 'body']):
            submission = Submission(self.reddit, 'n49rw')
            skipped = submission.comments.replace_more(None, threshold=0)
            assert len(skipped) == 0
            assert len(submission.comments.list()) > 1000
            assert len(submission.comments.list()) == \
                len(submission.comments._comments_by_id)

    def test_replace__all_with_comment_limit(self):
        with self.recorder.use_cassette(
                'TestCommentForest.test_replace__all_with_comment_limit',
                match_requests_on=['uri', 'method', 'body']):
            submission = Submission(self.reddit, '3hahrw')
            submission.comment_limit = 10
            skipped = submission.comments.replace_more(None, threshold=0)
            assert len(skipped) == 0
            assert len(submission.comments.list()) >= 500

    def test_replace__skip_at_limit(self):
        with self.recorder.use_cassette(
                'TestCommentForest.test_replace__skip_at_limit',
                match_requests_on=['uri', 'method', 'body']):
            submission = Submission(self.reddit, '3hahrw')
            skipped = submission.comments.replace_more(1)
            assert len(skipped) == 17

    def test_replace__skip_below_threshold(self):
        with self.recorder.use_cassette(
                'TestCommentForest.test_replace__skip_below_threshold',
                match_requests_on=['uri', 'method', 'body']):
            submission = Submission(self.reddit, '3hahrw')
            before_count = len(submission.comments.list())
            skipped = submission.comments.replace_more(16, 5)
            assert len(skipped) == 13
            assert all(x.count < 5 for x in skipped)
            assert all(x.submission == submission for x in skipped)
            assert before_count < len(submission.comments.list())

    def test_replace__skip_all(self):
        with self.recorder.use_cassette(
                'TestCommentForest.test_replace__skip_all',
                match_requests_on=['uri', 'method', 'body']):
            submission = Submission(self.reddit, '3hahrw')
            before_count = len(submission.comments.list())
            skipped = submission.comments.replace_more(limit=0)
            assert len(skipped) == 18
            assert all(x.submission == submission for x in skipped)
            after_count = len(submission.comments.list())
            assert before_count == after_count + len(skipped)



import pickle

import pytest
from praw.models import Redditor

from ... import UnitTest


class TestSubredit(UnitTest):
    def test_equality(self):
        redditor1 = Redditor(self.reddit, _data={'name': 'dummy1', 'n': 1})
        redditor2 = Redditor(self.reddit, _data={'name': 'Dummy1', 'n': 2})
        redditor3 = Redditor(self.reddit, _data={'name': 'dummy3', 'n': 2})
        assert redditor1 == redditor1
        assert redditor2 == redditor2
        assert redditor3 == redditor3
        assert redditor1 == redditor2
        assert redditor2 != redditor3
        assert redditor1 != redditor3
        assert 'dummy1' == redditor1
        assert redditor2 == 'dummy1'

    def test_construct_failure(self):
        message = 'Either `name` or `_data` must be provided.'
        with pytest.raises(TypeError) as excinfo:
            Redditor(self.reddit)
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Redditor(self.reddit, 'dummy', {'id': 'dummy'})
        assert str(excinfo.value) == message

    def test_fullname(self):
        redditor = Redditor(self.reddit, _data={'name': 'name', 'id': 'dummy'})
        assert redditor.fullname == 't2_dummy'

    def test_guild__min(self):
        with pytest.raises(TypeError) as excinfo:
            Redditor(self.reddit, name='RedditorName').gild(0)
        assert str(excinfo.value) == 'months must be between 1 and 36'

    def test_guild__max(self):
        with pytest.raises(TypeError) as excinfo:
            Redditor(self.reddit, name='RedditorName').gild(37)
        assert str(excinfo.value) == 'months must be between 1 and 36'

    def test_hash(self):
        redditor1 = Redditor(self.reddit, _data={'name': 'dummy1', 'n': 1})
        redditor2 = Redditor(self.reddit, _data={'name': 'Dummy1', 'n': 2})
        redditor3 = Redditor(self.reddit, _data={'name': 'dummy3', 'n': 2})
        assert hash(redditor1) == hash(redditor1)
        assert hash(redditor2) == hash(redditor2)
        assert hash(redditor3) == hash(redditor3)
        assert hash(redditor1) == hash(redditor2)
        assert hash(redditor2) != hash(redditor3)
        assert hash(redditor1) != hash(redditor3)

    def test_pickle(self):
        redditor = Redditor(self.reddit, _data={'name': 'name', 'id': 'dummy'})
        for level in range(pickle.HIGHEST_PROTOCOL + 1):
            other = pickle.loads(pickle.dumps(redditor, protocol=level))
            assert redditor == other

    def test_repr(self):
        redditor = Redditor(self.reddit, name='RedditorName')
        assert repr(redditor) == 'Redditor(name=\'RedditorName\')'

    def test_str(self):
        redditor = Redditor(self.reddit, _data={'name': 'name', 'id': 'dummy'})
        assert str(redditor) == 'name'



from praw.models import MoreComments

from ... import IntegrationTest


class TestMore(IntegrationTest):
    def setup(self):
        super(TestMore, self).setup()
        # Responses do not decode well on travis so manually renable gzip.
        self.reddit._core._requestor._http.headers['Accept-Encoding'] = 'gzip'

    def test_comments(self):
        data = {'count': 9, 'name': 't1_cu5tt8h', 'id': 'cu5tt8h',
                'parent_id': 't3_3hahrw', 'children': [
                    'cu5tt8h', 'cu5v9yd', 'cu5twf5', 'cu5tkk4', 'cu5tead',
                    'cu5rxpy', 'cu5oufs', 'cu5tpek', 'cu5pbdh']}
        with self.recorder.use_cassette(
                'TestMore.test_comments',
                match_requests_on=['uri', 'method', 'body']):
            more = MoreComments(self.reddit, data)
            more.submission = self.reddit.submission('3hahrw')
            assert len(more.comments()) == 7

    def test_comments__continue_thread_type(self):
        data = {'count': 0, 'name': 't1__', 'id': '_',
                'parent_id': 't1_cu5v5h7', 'children': []}
        with self.recorder.use_cassette(
                'TestMore.test_comments__continue_thread_type',
                match_requests_on=['uri', 'method', 'body']):
            more = MoreComments(self.reddit, data)
            more.submission = self.reddit.submission('3hahrw')
            assert len(more.comments()) == 1



import pickle

from praw.models import MoreComments

from ... import UnitTest


class TestComment(UnitTest):
    def test_pickle(self):
        more = MoreComments(self.reddit, {'children': ['a', 'b'], 'count': 4})
        for level in range(pickle.HIGHEST_PROTOCOL + 1):
            other = pickle.loads(pickle.dumps(more, protocol=level))
            assert more == other

    def test_repr(self):
        more = MoreComments(self.reddit, {
            'children': ['a', 'b', 'c', 'd', 'e'], 'count': 5})
        assert repr(more) == ("<MoreComments count=5, "
                              "children=['a', 'b', 'c', '...']>")

        more = MoreComments(self.reddit, {
            'children': ['a', 'b', 'c', 'd'], 'count': 4})
        assert repr(more) == ("<MoreComments count=4, "
                              "children=['a', 'b', 'c', 'd']>")



"""Provide the SubredditListingMixin class."""
from six.moves.urllib.parse import urljoin  # pylint: disable=import-error

from ..generator import ListingGenerator
from .base import BaseListingMixin
from .gilded import GildedListingMixin


class SubredditListingMixin(BaseListingMixin, GildedListingMixin):
    """Adds additional methods pertianing to Subreddit-like instances."""

    def __init__(self, reddit, _data):
        """Initialize a SubredditListingMixin instance.

        :param reddit: An instance of :class:`.Reddit`.

        """
        super(SubredditListingMixin, self).__init__(reddit, _data)
        self.comments = CommentHelper(self)

    def rising(self, **generator_kwargs):
        """Return a ListingGenerator for rising submissions.

        Additional keyword arguments are passed to the ``ListingGenerator``
        constructor.

        """
        return ListingGenerator(self._reddit, urljoin(self._path, 'rising'),
                                **generator_kwargs)


class CommentHelper(GildedListingMixin):
    """Provide a set of functions to interact with a subreddit's comments."""

    @property
    def _path(self):
        return urljoin(self.subreddit._path, 'comments/')

    def __init__(self, subreddit):
        """Initialize a CommentHelper instance."""
        super(CommentHelper, self).__init__(subreddit._reddit, None)
        self.subreddit = subreddit

    def __call__(self, **generator_kwargs):
        """Return a ListingGenerator for the Subreddit's comments."""
        return ListingGenerator(self._reddit, self._path, **generator_kwargs)



"""Provide the Redditor class."""
from json import dumps

from ...const import API_PATH
from ..listing.mixins import RedditorListingMixin
from .base import RedditBase
from .mixins import MessageableMixin


class Redditor(RedditBase, MessageableMixin, RedditorListingMixin):
    """A class representing the users of reddit."""

    STR_FIELD = 'name'

    @classmethod
    def from_data(cls, reddit, data):
        """Return an instance of Redditor, or None from ``data``."""
        if data == '[deleted]':
            return None
        else:
            return cls(reddit, data)

    def __init__(self, reddit, name=None, _data=None):
        """Initialize a Redditor instance.

        :param reddit: An instance of :class:`~.Reddit`.
        :param name: The name of the redditor.

        """
        if bool(name) == bool(_data):
            raise TypeError('Either `name` or `_data` must be provided.')
        super(Redditor, self).__init__(reddit, _data)
        self._listing_use_sort = True
        if name:
            self.name = name
        self._path = API_PATH['user'].format(user=self)

    def _info_path(self):
        return API_PATH['user_about'].format(user=self)

    def _friend(self, method, data):
        url = API_PATH['friend_v1'].format(user=self)
        return self._reddit.request(method, url, data=dumps(data))

    def friend(self, note=None):
        """Friend the Redditor.

        :param note: A personal note about the user. Requires reddit
            Gold. (Default: None)
        :returns: The json response from the server.

        Calling this method subsequent times will update the note.

        """
        return self._friend('PUT', data={'note': note} if note else {})

    def friend_info(self):
        """Return a Redditor instance with specific friend-related attributes.

        :returns: A Redditor instance with fields ``date``, ``id``, and
            possibly ``note`` if the authenticated user has reddit Gold.

        """
        return self._reddit.get(API_PATH['friend_v1'].format(user=self))

    def gild(self, months=1):
        """Gild the Redditor.

        :param months: Specifies the number of months to gild up to 36
            (default: 1).

        """
        if months < 1 or months > 36:
            raise TypeError('months must be between 1 and 36')
        self._reddit.post(API_PATH['gild_user'].format(username=self),
                          data={'months': months})

    def multireddits(self):
        """Return a list of the redditor's public multireddits."""
        return self._reddit.get(API_PATH['multireddit_user'].format(user=self))

    def unblock(self):
        """Unblock the Redditor.

        :returns: The json response from the server.

        Blocking must be done from a Message, Comment Reply or Submission
        Reply.

        """
        data = {'container': self._reddit.user.me().fullname,
                'name': str(self), 'type': 'enemy'}
        return self._reddit.post(API_PATH['unfriend'], data=data)

    def unfriend(self):
        """Unfriend the Redditor."""
        self._friend(method='delete', data={'id': str(self)})



"""praw constants."""

__version__ = '4.0.0b13'

API_PATH = {
    'accept_mod_invite':      'api/accept_moderator_invite',
    'access_token_url':       'api/v1/access_token/',
    'approve':                'api/approve/',
    'authorize':              'api/v1/authorize/',
    'banned':                 'r/{subreddit}/about/banned/',
    'block':                  'api/block',
    'blocked':                'prefs/blocked/',
    'by_id':                  'by_id/',
    'captcha':                'captcha/',
    'clearflairtemplates':    'api/clearflairtemplates/',
    'collapse_message':       'api/collapse_message/',
    'comment':                'api/comment/',
    'comment_replies':        'message/comments/',
    'comments':               'comments/',
    'compose':                'api/compose/',
    'contest_mode':           'api/set_contest_mode/',
    'contributor':            'r/{subreddit}/about/contributors/',
    'controversial':          'controversial/',
    'del':                    'api/del/',
    'deleteflair':            'api/deleteflair',
    'delete_redditor':        'api/delete_user',
    'delete_sr_header':       'r/{subreddit}/api/delete_sr_header',
    'delete_sr_image':        'r/{subreddit}/api/delete_sr_img',
    'distinguish':            'api/distinguish/',
    'domain':                 'domain/{domain}/',
    'duplicates':             'duplicates/{submission_id}/',
    'edit':                   'api/editusertext/',
    'edited':                 'r/{subreddit}/about/edited/',
    'flair':                  'r/{subreddit}/api/flair/',
    'flairconfig':            'api/flairconfig/',
    'flaircsv':               'r/{subreddit}/api/flaircsv/',
    'flairlist':              'r/{subreddit}/api/flairlist/',
    'flairselector':          'r/{subreddit}/api/flairselector/',
    'flairtemplate':          'api/flairtemplate/',
    'friend':                 'api/friend/',
    'friend_v1':              'api/v1/me/friends/{user}',
    'friends':                'api/v1/me/friends/',
    'gild_thing':             'api/v1/gold/gild/{fullname}/',
    'gild_user':              'api/v1/gold/give/{username}/',
    'help':                   'help/',
    'hide':                   'api/hide/',
    'ignore_reports':         'api/ignore_reports/',
    'inbox':                  'message/inbox/',
    'info':                   'api/info/',
    'leavecontributor':       'api/leavecontributor',
    'leavemoderator':         'api/leavemoderator',
    'liveabout':              'api/live/{id}/about/',
    'livecreate':             'api/live/create',
    'me':                     'api/v1/me',
    'mentions':               'message/mentions',
    'message':                'message/messages/{messageid}/',
    'messages':               'message/messages/',
    'moderator':              'r/{subreddit}/about/moderators/',
    'moderator_messages':     'r/{subreddit}/message/moderator/',
    'moderator_unread':       'r/{subreddit}/message/moderator/unread/',
    'modlog':                 'r/{subreddit}/about/log/',
    'modqueue':               'r/{subreddit}/about/modqueue/',
    'morechildren':           'api/morechildren/',
    'my_contributor':         'subreddits/mine/contributor/',
    'my_moderator':           'subreddits/mine/moderator/',
    'my_multireddits':        'api/multi/mine/',
    'my_subreddits':          'subreddits/mine/subscriber/',
    'new':                    'new/',
    'marknsfw':               'api/marknsfw/',
    'multireddit':            'user/{user}/m/{multi}/',
    'multireddit_api':        'api/multi/user/{user}/m/{multi}/',
    'multireddit_base':       'api/multi/',
    'multireddit_copy':       'api/multi/copy/',
    'multireddit_rename':     'api/multi/rename/',
    'multireddit_update':     'api/multi/user/{user}/m/{multi}/r/{subreddit}',
    'multireddit_user':       'api/multi/user/{user}/',
    'mute_sender':            'api/mute_message_author/',
    'muted':                  'r/{subreddit}/about/muted/',
    'read_message':           'api/read_message/',
    'reddit_url':             '/',
    'register':               'api/register/',
    'remove':                 'api/remove/',
    'report':                 'api/report/',
    'reports':                'r/{subreddit}/about/reports/',
    'rising':                 'rising/',
    'save':                   'api/save/',
    'saved':                  'saved/',
    'search':                 'r/{subreddit}/search/',
    'select_flair':           'r/{subreddit}/api/selectflair/',
    'sent':                   'message/sent/',
    'sticky':                 'r/{subreddit}/about/sticky/',
    'sticky_submission':      'api/set_subreddit_sticky/',
    'site_admin':             'api/site_admin/',
    'spam':                   'r/{subreddit}/about/spam/',
    'stylesheet':             'r/{subreddit}/about/stylesheet/',
    'sub_recommendations':    'api/recommend/sr/{subreddits}',
    'submission':             'comments/{id}/',
    'submission_replies':     'message/selfreply/',
    'submit':                 'api/submit/',
    'subreddit':              'r/{subreddit}/',
    'subreddit_about':        'r/{subreddit}/about/',
    'subreddit_comments':     'r/{subreddit}/comments/',
    'subreddit_css':          'api/subreddit_stylesheet/',
    'subreddit_random':       'r/{subreddit}/random/',
    'subreddit_settings':     'r/{subreddit}/about/edit/',
    'subreddit_traffic':      'r/{subreddit}/about/traffic/',
    'subreddits_default':     'subreddits/default/',
    'subreddits_gold':        'subreddits/gold/',
    'subreddits_new':         'subreddits/new/',
    'subreddits_popular':     'subreddits/popular/',
    'subreddits_name_search': 'api/search_reddit_names/',
    'subreddits_search':      'subreddits/search/',
    'subscribe':              'api/subscribe/',
    'suggested_sort':         'api/set_suggested_sort/',
    'top':                    'top/',
    'uncollapse_message':     'api/uncollapse_message/',
    'unfriend':               'api/unfriend/',
    'unhide':                 'api/unhide/',
    'unmarknsfw':             'api/unmarknsfw/',
    'unmoderated':            'r/{subreddit}/about/unmoderated/',
    'unmute_sender':          'api/unmute_message_author/',
    'unignore_reports':       'api/unignore_reports/',
    'unread':                 'message/unread/',
    'unread_message':         'api/unread_message/',
    'unsave':                 'api/unsave/',
    'upload_image':           'api/upload_sr_img',
    'user':                   'user/{user}/',
    'user_about':             'user/{user}/about/',
    'username_available':     'api/username_available/',
    'vote':                   'api/vote/',
    'wiki_edit':              'r/{subreddit}/api/wiki/edit/',
    'wiki_page':              'r/{subreddit}/wiki/{page}',
    'wiki_page_editor':       'r/{subreddit}/api/wiki/alloweditor/{method}',
    'wiki_page_settings':     'r/{subreddit}/wiki/settings/{page}',
    'wiki_pages':             'r/{subreddit}/wiki/pages/',
    'wikibanned':             'r/{subreddit}/about/wikibanned/',
    'wikicontributor':        'r/{subreddit}/about/wikicontributors/'}

JPEG_HEADER = b'\xff\xd8\xff'
MAX_IMAGE_SIZE = 512000
MIN_PNG_SIZE = 67
MIN_JPEG_SIZE = 128
PNG_HEADER = b'\x89\x50\x4e\x47\x0d\x0a\x1a\x0a'

USER_AGENT_FORMAT = '{{}} PRAW/{}'.format(__version__)



"""PRAW Integration test suite."""
import pytest
from betamax import Betamax
from praw import Reddit


class IntegrationTest(object):
    """Base class for PRAW integration tests."""

    def setup(self):
        """Setup runs before all test cases."""
        self.setup_reddit()
        self.setup_betamax()

    def setup_betamax(self):
        """Configure betamax instance based off of the reddit instance."""
        http = self.reddit._core._requestor._http
        self.recorder = Betamax(http)

        # Disable response compression in order to see the response bodies in
        # the betamax cassettes.
        http.headers['Accept-Encoding'] = 'identity'

        # Require tests to explicitly disable read_only mode.
        self.reddit.read_only = True

    def setup_reddit(self):
        self.reddit = Reddit(client_id=pytest.placeholders.client_id,
                             client_secret=pytest.placeholders.client_secret,
                             password=pytest.placeholders.password,
                             user_agent=pytest.placeholders.user_agent,
                             username=pytest.placeholders.username)


class UntrustedIntegrationTest(IntegrationTest):
    def setup_reddit(self):
        self.reddit = Reddit(client_id=pytest.placeholders.client_id,
                             client_secret=None,
                             user_agent=pytest.placeholders.user_agent)



import os
import sys

import mock
import pytest
from praw.config import Config
from praw.exceptions import ClientException


class TestConfig(object):
    @staticmethod
    def _assert_config_read(environment, mock_config):
        mock_instance = mock_config.return_value
        Config.CONFIG = None  # Force config file reload
        prev_environment = {environment: None}
        for env_name in ['APPDATA', 'HOME', 'XDG_CONFIG_HOME']:
            if env_name in os.environ:
                prev_environment[env_name] = os.environ[env_name]
                del os.environ[env_name]
        os.environ[environment] = '/MOCK'

        module_dir = os.path.dirname(sys.modules['praw'].__file__)
        environ_path = os.path.join(
            '/MOCK', '.config' if environment == 'HOME' else '', 'praw.ini')
        locations = [os.path.join(module_dir, 'praw.ini'), environ_path,
                     'praw.ini']

        try:
            Config._load_config()
            mock_instance.read.assert_called_with(locations)
        finally:
            Config.CONFIG = None  # Force config file reload
            for env_name in prev_environment:
                if prev_environment[env_name] is None:
                    del os.environ[env_name]
                else:
                    os.environ[env_name] = prev_environment[env_name]

    @mock.patch('six.moves.configparser.RawConfigParser')
    def test_load_ini_from_appdata(self, mock_config):
        self._assert_config_read('APPDATA', mock_config)

    @mock.patch('six.moves.configparser.RawConfigParser')
    def test_load_ini_from_home(self, mock_config):
        self._assert_config_read('HOME', mock_config)

    @mock.patch('six.moves.configparser.RawConfigParser')
    def test_load_ini_from_xdg_config_home(self, mock_config):
        self._assert_config_read('XDG_CONFIG_HOME', mock_config)

    @mock.patch('six.moves.configparser.RawConfigParser')
    def test_load_ini_with_no_config_directory(self, mock_config):
        mock_instance = mock_config.return_value
        Config.CONFIG = None  # Force config file reload

        prev_environment = {}
        for key in ['APPDATA', 'HOME', 'XDG_CONFIG_HOME']:
            if key in os.environ:
                prev_environment[key] = os.environ[key]
                del os.environ[key]

        module_dir = os.path.dirname(sys.modules['praw'].__file__)
        locations = [os.path.join(module_dir, 'praw.ini'), 'praw.ini']

        try:
            Config._load_config()
            mock_instance.read.assert_called_with(locations)
        finally:
            Config.CONFIG = None  # Force config file reload
            for key, value in prev_environment.items():
                os.environ[key] = value

    def test_short_url(self):
        config = Config('DEFAULT')
        assert config.short_url == 'https://redd.it'

    def test_short_url_not_defined(self):
        config = Config('DEFAULT', short_url=None)
        with pytest.raises(ClientException) as excinfo:
            config.short_url
        assert str(excinfo.value) == 'No short domain specified.'
